<?php
// +----------------------------------------------------------------------
// Swoole - WebSocket 长连接
// +----------------------------------------------------------------------
// Copyright (c) 2018 https://xiuxian.junphp.com All rights reserved.
// +----------------------------------------------------------------------
// Licensed ( http://www.apache.org/licenses/LICENSE-2.0 )
// +----------------------------------------------------------------------
// Author: 小黄牛 <1731223728@qq.com>
// +----------------------------------------------------------------------
/**
 * CODE:
 * 1 上线
 * 2 心跳重连
 * 3 下线
 * 4 消息广播
 * 5 强制下线
 * 6 心跳检测
 * 7 查询渠道信息
 * 8 查询渠道下的用户列表
*/
class Server{
	/**
	 * WS的启动实例
	*/
	private $_ws;
	/**
	 * host-IP，0.0.0.0表示允许接收所有请求
	*/
	private $_host = '0.0.0.0';
	/**
	 * 端口号
	*/
	private $_port = '9502';
	/**
	 * 最大服务端心跳重连次数
	*/
	private $_max  = 3;
	/**
	 * 强制心跳重连启动状态
	*/
	private $_status = true;
	/**
	 * 数据库配置
	*/
	private $_config = [];
	/**
	 * 数据库实例
	*/
	private $DB;

    /**
     * 这是启动服务端的入口
    */
    public function run() { 
		$this->_config = require_once dirname(dirname(__FILE__)).'/config/database.php';
		
		// 打开数据库
		try {
			$dsn      = $this->_config['type'].':dbname='.$this->_config['database'].';host='.$this->_config['hostname'].';port='.$this->_config['hostport'];
			$this->DB = new PDO($dsn, $this->_config['username'], $this->_config['password']);
		} catch(PDOException $e) {
			die ('连接失败，错误信息：'.$e->getMessage());
		}
		$this->DB->query('SET NAMES '.$this->_config['charset'].';');

		$this->start_service(); 
		$this->start_table(); 
		$this->request();
		$this->start_handshake();
		$this->start_message();
		$this->end();
    }

	/**
	 * ①启动websocker服务
	*/
	private function start_service() {
		# 创建websocket服务器对象，监听0.0.0.0:9502端口
		$this->_ws = new swoole_websocket_server($this->_host, $this->_port);
		# 根据CPU，自动识别进程数
		// $this->_ws->set([
        //     'worker_num' => 4,// 开4个工作进程
		// ]);
	}

	/**
	 * ①创建Table服务
	*/
	private function start_table() {
		# 创建最大只能存储1024个渠道商的表
		$channel = new swoole_table(1024);
		# 创建字段
		$channel->column('secret_key', swoole_table::TYPE_STRING, 255); // 密钥
		$channel->column('max_num', swoole_table::TYPE_INT, 11); // 最大连接数
		$channel->column('on_line', swoole_table::TYPE_INT, 11); // 在线用户数
		$channel->column('status', swoole_table::TYPE_INT, 2); // 状态
		$channel->create();
		$this->_ws->channel = $channel;

		# 查询所有加盟商
		$res  = $this->DB->query('SELECT * FROM '.$this->_config['prefix'].'channel;');
		$list = $res->FetchAll();
		# 最大在线用户量，默认10，防止渠道商为空时启动失败
		$max_num = 10;
		foreach ($list as $v) {
			# 加入存储器
			$this->_ws->channel->set($v['appid'], [
				'secret_key' => $v['secret_key'],
				'max_num'    => $v['max_num'],
				'status'     => $v['status'],
				'on_line'    => 0,
			]);
			$max_num += $v['max_num'];
		}

		# 创建用户表
		$user = new swoole_table($max_num);
		# 创建字段
		$user->column('fd', swoole_table::TYPE_INT, 8); // FD
		$user->column('status', swoole_table::TYPE_INT, 8); // 离线状态
		$user->column('heartbeat', swoole_table::TYPE_INT, 8); // 心跳重连数
		$user->column('id', swoole_table::TYPE_STRING, 32); // 会员ID
		$user->column('appid', swoole_table::TYPE_STRING, 128); // APPID
		$user->create();
		$this->_ws->user = $user;
	}

	/**
	 * ②监听WebSocket握手申请
	*/
	private function start_handshake() {
		# 监听WebSocket连接打开事件
		$this->_ws->on('open', function ($ws, $request){
			# 鉴权
			$param = $request->get;
			$this->Auth($param, $request->fd);
		});
	}

	/**
	 * 身份鉴权
	*/
	private function Auth($param, $fd) {
		# 非法渠道直接断开
		if (empty($param['appid'])) {
			$this->_ws->close($fd); 
			return false;
		}
		# 非法渠道直接断开
		$channel = $this->_ws->channel->get($param['appid']);
		if (!$channel) {
			$this->_ws->close($fd);
			return false;
		}
		# 禁用直接关闭
		if ($channel['status']==0) {
			$this->_ws->close($fd);
			return false;
		}
		return $channel;
	}

	/**
	 * ③监听客户端消息发送请求
	*/
	private function start_message() {
		# 监听WebSocket消息事件
		$this->_ws->on('message', function ($ws, $frame) {
			$data    = json_decode($frame->data, true);
			$channel = $this->Auth($data, $frame->fd);
			$user_id = $data['user_id'];

			# 加入存储器
			$this->_ws->user->set($user_id, [
				'appid'		=> $data['appid'],#APPID
				'fd'        => $frame->fd, # FD
				'status'    => 1, # 设置上线状态
				'heartbeat' => 0, # 重置心跳重连数
				'id' 		=> $data['user_id'], # 用户ID
			]);
			# 在线人数统计+1
			$this->_ws->channel->set($data['appid'], [
				'on_line' => $channel['on_line']+1,
			]);
			
			# 上线通知
			if ($data['code'] == 1) {
				$this->broadcast($ws, $data, $user_id);
			# 心跳重连检测
			} else if ($data['code'] == 2) {
				$this->broadcast($ws, $data, $user_id);
				$this->timer($data['appid']);
			# 其他请求
			} else {
				# 广播消息
				$this->broadcast($ws, $data, $user_id);
			}
		});
	}

	/**
	 * ④监听客户端退出事件
	*/
	private function end() {
		# 这里加入了unset，清除open存储器，防止存储器无限增大
		# 监听WebSocket连接关闭事件
		$this->_ws->on('close', function ($ws, $fd) {
			$user = null;
			foreach ($this->_ws->user as $k=>$v) {
				if ($v['fd'] == $fd) {
					$user = $v;
				}

			}
			# 如果没用用户就跳过
			if (!$user) {
				return false;
			}

			# 获取用户ID
			$user_id = $user['id'];
			# 直接删除
			$this->_ws->user->del($user_id);
			# 更新在线人数
			$channel = $this->_ws->channel->get($user['appid']);
			$this->_ws->channel->set($user['appid'], [
				'on_line' => $channel['on_line']-1,
			]);

			$data = [
				'code' => 3,
				'user_id' => $user_id,
			];
			# 广播消息
			$this->broadcast($ws, $data, null, false, $user['appid']);
		});
		$this->_ws->start();
	}

	/**
	 * ④监听外部请求推送事件
	*/
	private function request() {
		$this->_ws->on('request', function ($request, $response) {
			//var_dump($request);
			# 如果你是get的，就改成get，可以用dump看看$request
			$param = $request->post;
			$data  = [];
			$data['code'] = $param['code'];
			if (!empty($param['user_id'])) {
				$data['user_id'] = $param['user_id'];
			}
			if (empty($param['appid'])) {
				# 返回消息
				$this->endRequest('500', '请输入APPID', [], $request, $response);
			}
			$channel = $this->_ws->channel->get($param['appid']);
			if (!$channel) {
				# 返回消息
				$this->endRequest('500', 'APPID错误', [], $request, $response);
			}
			if ($channel['status']==0) {
				# 返回消息
				$this->endRequest('500', '该渠道已被禁用', [], $request, $response);
			}

			# 查询渠道信息
			if ($data['code'] == 7) {
				# 返回消息
				$this->endRequest('200', '获取成功', $channel, $request, $response);
			# 查询渠道下的用户列表
			} else if ($data['code'] == 8) {	
				$list = [];
				foreach ($this->_ws->user as $k=>$v) {
					if ($v['appid'] == $param['appid']) {
						$list[] = $v;
					}
				}
				$list = $this->ArrayPage($list, $param['limit'], $param['page']);
				# 返回消息
				$this->endRequest('200', '获取成功', $list, $request, $response);
			# 下面我们来广播消息
			} else if (empty($param['user_id'])) {
				# 群发
				$this->broadcast($this->_ws, $data, '', false, $param['appid']);
				# 返回消息
				$this->endRequest('200', '发送成功', [], $request, $response);
			} else {
				# 单发
				$user = $this->_ws->user->get($param['user_id']);
				if (empty($user)) {
					# 返回消息
					$this->endRequest('500', '客户不存在', [], $request, $response);
				} else {
					if ($user['status'] == 0) {
						# 返回消息
						$this->endRequest('500', '客户已下线', [], $request, $response);
					} else {
						$this->_ws->push($user['fd'], $this->json($data));
						# 返回消息
						$this->endRequest('200', '发送成功', [], $request, $response);
					}
				}
			}
		});
	}

	/**
	 * 数组分页
	*/
	private function ArrayPage($array,$size,$page){
		$num =count($array);//总长度
		$pagenum = ceil($num / $size);//总页数
		if($page==0 && $page==''){
			$page = 1;
		}else{
			$page = $page?$page:1;
		}
		$arr = array();
		for($i = 0; $i < $size; $i++){
			$a = @$array[($page - 1) * $size + $i];
			if ($a) $arr[] = $a;
			
		}
		return $arr;
	}

	/**
	 * request事件返回值
	*/
	private function endRequest($code, $msg, $data, $request, $response) {
		$json = [
			'code' => "$code",
			'msg'  => "$msg",
			'data'  => $data,
		];
		# 输出响应
		$return = json_encode($json, JSON_UNESCAPED_UNICODE);
		# 输出返回值
		$response->end($return);
	}

	/**
	 * 广播消息
	 * @todo 无
	 * @author 小黄牛
	 * @version v1.0.0.1 + 2018.11.12
	 * @deprecated 暂不弃用
	 * @global 无
	 * @param object $wx 实例
	 * @param string $content 广播内容
	 * @param string $id 用户的userid
	 * @param bool $status 是否做心跳限制 
	 * @param string $appid 渠道ID 
	 * @return void
	*/
	private function broadcast($ws, $content, $id=null, $status=false, $appid=null) {
		# 向所有人广播
		if ($id) {
			$user = $this->_ws->user->get($id);
			$appid = $user['appid'];
		}
		foreach ($this->_ws->user as $k=>$v) {
			# 只能给渠道内的人员发送消息
			# 不向自己广播，并且要在线的
			# 注意，这里一定要有上线状态的限制，否则假设用户已经退出，但你的进程还开着，实际上已经关闭，这时候push就会报错
			# 只有正常在线的用户才能接收到广播
			# 加入心跳检测限制
			if ($v['appid'] == $appid && $k != $id && $v['status'] == 1 && $status == true) {
				$ws->push($v['fd'], $this->json($content));
			} else if ($v['appid'] == $appid && $k != $id && $v['status'] == 1 && $v['heartbeat'] == 0) {
				$ws->push($v['fd'], $this->json($content));
			}
		}
	}

	/**
	 * 数组转json
	 * @todo 无
	 * @author 小黄牛
	 * @version v1.0.0.1 + 2018.11.08
	 * @deprecated 暂不弃用
	 * @global 无
	 * @param array $array 数组
	 * @return json
	*/
	private function json($array) {
		return json_encode($array, JSON_UNESCAPED_UNICODE);
	}

	/**
	 * 服务端定时强制心跳检测
	 * @todo 无
	 * @author 小黄牛
	 * @version v1.0.0.1 + 2018.11.08
	 * @deprecated 暂不弃用
	 * @global 无
	 * @param string $appid 渠道ID 
	 * @return void
	*/
	private function timer($appid) {
		# 注意强制心跳触发器不能放在open事件里，因为那时候用户还没有提交登录请求，是还没有userID的
		# 还有，强制心跳定时器只能触发一次，否则会出现生成多个定时器的情况
		if ($this->_status) {
			$this->_status = false;
			/**
			 * ⑤服务端强制心跳检测
			 * 每隔1分钟发送1次，如果连续3次强制心跳检测未通过，服务端将强制断开连接
			*/
			$obj = $this;
			swoole_timer_tick(60000, function ($timer_id) use (&$obj, &$appid) {
				# 广播消息
				$obj->broadcast($obj->_ws, ['code' => 6], null, true, $appid);
				# 所有人的心跳次数+1
				foreach ($this->_ws->user as $k=>$v) {
					if (empty($v['heartbeat'])) {
						# 重置心跳次数
						$this->_ws->user->set($v['id'], [
							'heartbeat' => 0,
						]);
					}
					# 心跳次数累加
					$this->_ws->user->set($v['id'], [
						'heartbeat' => $v['heartbeat']+1
					]);
					# 心跳次数大于等于_max && 在线的 的连接关闭
					if ($v['heartbeat'] >= $obj->_max && $v['status'] == 1) {
						$data = $v;
						# 发送强制掉线广播
						$data['code'] = 5;
						$obj->broadcast($obj->_ws, $data, null, true, $appid);
						# 这里不需要unset连接，因为在close事件中，已经将这个连接设置为离线了
						# 主动关闭连接k
						$obj->_ws->close($v['fd']);
					}
				}
			});
		}
	}
}

$socketServer = new Server();
$socketServer->run();